#ifdef USE_CAPI20

#include "IsdnCapiConnection.h"
#include "IsdnCapiConnectionManager.h"
#include "IsdnCapiUtil.h"
#include "IsdnConfiguration.h"
#include "IsdnUtils.h"
#include "AmApi.h"
#include "globals.h"
#include "log.h"
#include <stdio.h>
#include <capi20.h>
#include <pthread.h>
#include <string.h>

IsdnCapiConnection::IsdnCapiConnection(int controller) {
	init();
	m_controller = controller;
}

IsdnCapiConnection::IsdnCapiConnection() {
	init();
}

void IsdnCapiConnection::init() {
	state = CC_STATE_IDLE;
	id = capiManager->createCCId();
	/* init the audio buffers for in&out */
	fi_readpos = fi_audiobuffer;
	fi_datalen = 0;
	ti_readpos = ti_audiobuffer;
	ti_datalen = 0;
	pthread_mutex_init(&fi_m, NULL);
	pthread_mutex_init(&ti_m, NULL);
	pthread_mutex_init(&callsetup_m, NULL);
	DBG("mutexes initialized");
}

IsdnCapiConnection::~IsdnCapiConnection() {
}

void IsdnCapiConnection::onBeforeCallAccept(AmRequest* req) {
	int ret;
	char phonenumber[MAX_NUM_LEN];
	char msn[MAX_NUM_LEN];
	int shownumber;
	
	shownumber = isdn_config->clip?1:0;

	IsdnUtils::determineNumbers(req, phonenumber, msn, MAX_NUM_LEN); /* throws Exception on error */
	INFO("determine numbers result: from=%s to=%s\n", phonenumber, msn);
	req->reply(183, "isdngw: call progress");
	
	/* lock the call setup mutex */
	DBG("locking call setup \n");
	pthread_mutex_lock (&callsetup_m);
	
	if (strlen(msn) != 0)
		ret = call(msn, phonenumber, shownumber);
	else
		ret = call(NULL, phonenumber, shownumber);
		
	if (ret == FAIL) {
		ERROR("capi call setup failed\n");
		throw AmSession::Exception(500,"isdngw: capi call setup failed");
	}

	req->reply(180, "isdngw: capi ringing");
	/* now capi call setup is in progress, check in intervals */
	DBG("start waiting on call setup mutex\n");
	pthread_mutex_lock(&callsetup_m);
	pthread_mutex_unlock(&callsetup_m);

	if (state == CC_STATE_CONNECTED) {
		DBG("done waiting. now connected\n");
		return;
	} else if (state == CC_STATE_WASTE) {
		DBG("done waiting. waste. code=%i reason=%s\n", m_code, m_reason);
		throw AmSession::Exception(m_code, m_reason);
	} else if (state == CC_STATE_DISCONNECTED) {
		DBG("done waiting. disconnected\n");
		throw AmSession::Exception(487, "call terminated");
		return;
	} else {
		DBG("done waiting. unknown state %i\n", state);
		throw AmSession::Exception(500,"isdngw: internal error (unknown state).");
	}
}
	
int IsdnCapiConnection::call(char* caller, char* callee, int shownumber) {
	_cmsg msg;
	unsigned char callerstruct[NUM_BUF_LEN];
	unsigned char calleestruct[NUM_BUF_LEN];
	int res;
	int ntype;

	DBG("capi call from %s to %s\n", caller, callee);

	state = CC_STATE_CONNECTING;

	CONNECT_REQ_HEADER(&msg, capiManager->getAppId(), 1, m_controller);

	if (caller == NULL) {
		caller = isdn_config->out_msn;
	};

	ntype = shownumber==0?TYPE_CALLER_SUPPRESS:TYPE_CALLER;
	if (IsdnCapiUtil::build_number(ntype, callerstruct, NUM_BUF_LEN, caller) == FAIL) {
		ERROR("error building origin number\n");
		return -1;
	}	

	if (IsdnCapiUtil::build_number(TYPE_CALLEE, calleestruct, NUM_BUF_LEN, callee) == FAIL) {
		ERROR("error building target number\n");
		return -1;
	}	

	DBG("IsdnCapiConnection::call: from %.*s to %.*s\n", (*calleestruct)-1, calleestruct+2, (*callerstruct)-2, callerstruct+3);

	CONNECT_REQ_CALLEDPARTYNUMBER(&msg)=calleestruct;
	CONNECT_REQ_CALLINGPARTYNUMBER(&msg)=callerstruct;

	CONNECT_REQ_CALLEDPARTYSUBADDRESS(&msg)=0;
	CONNECT_REQ_CALLINGPARTYSUBADDRESS(&msg)=0;
	
	CONNECT_REQ_CIPVALUE(&msg)=16; /* accept speech */

	CONNECT_REQ_B1PROTOCOL(&msg)=1;
	CONNECT_REQ_B2PROTOCOL(&msg)=1;
	CONNECT_REQ_B3PROTOCOL(&msg)=0;
	
	CONNECT_REQ_B1CONFIGURATION(&msg)=NULL;
	CONNECT_REQ_B2CONFIGURATION(&msg)=NULL;
	CONNECT_REQ_B3CONFIGURATION(&msg)=NULL;
	
	msg.Messagenumber = id;
	
	capiManager->registerOutgoingCall(id, this);
	
	res = IsdnCapiUtil::putMessage(&msg);

	if (res == FAIL) {
		ERROR("ERROR: could not put message connect_req\n");
		return FAIL;
	}
		
	return OK;
}

int IsdnCapiConnection::cb_connect_conf(_cmsg* msg) {
	DBG("callback funtion called: cb_connect_conf!\n");
	capiManager->updateOutgoingCall(id, msg->adr.adrPLCI, this);
	adr = msg->adr.adrPLCI;
	return OK;
}

int IsdnCapiConnection::cb_connect_ind(_cmsg* msg) {
	DBG("callback funtion called: cb_connect_ind! (unimplemented)\n");
	return OK;
}

int IsdnCapiConnection::cb_connect_active_ind(_cmsg* msg) {
	int res;
	_cmsg msg2;
	DBG("callback funtion called: cb_connect_active_ind!\n");
	capi20_cmsg_answer(msg);
	res = IsdnCapiUtil::putMessage(msg);
	if (res == FAIL) {
		ERROR("could not put message connect_active_response\n");
		return FAIL;
	}

	DBG("physical connection established, now creating logical connection\n");
	CONNECT_B3_REQ_HEADER(&msg2, capiManager->getAppId(), id, msg->adr.adrPLCI);
	res = IsdnCapiUtil::putMessage(&msg2);
	if (res == FAIL) {
		ERROR("could not put message connect_b3_req\n");
		return FAIL;
	}

	return OK;
}

int IsdnCapiConnection::cb_connect_b3_conf(_cmsg* msg) {
	capiManager->updateOutgoingCall(id, msg->adr.adrPLCI, this);
	adr = msg->adr.adrPLCI;
	DBG("connected, now unlocking mutex\n");
	state = CC_STATE_CONNECTED;
	pthread_mutex_unlock(&callsetup_m);
	return OK;
}

int IsdnCapiConnection::cb_connect_b3_active_ind(_cmsg* msg) {
	int res;
	capi20_cmsg_answer(msg);
	res = IsdnCapiUtil::putMessage(msg);
	if (res == FAIL) {
		ERROR("could not put message connect_active_response\n");
		return FAIL;
	}
	return OK;
}

int IsdnCapiConnection::cb_disconnect_ind(_cmsg* msg) {
	int reason;
	int res;
	DBG("IsdnCapiConnection::cb_disconnect_ind\n");
	reason = DISCONNECT_IND_REASON(msg);
	if (reason == 0x349c) {
		ERROR("END CALL: invalid number format\n");
		m_code = 500;
		strcpy(m_reason,"isdngw: invalid number format (capi error: 0x349c)");
		state = CC_STATE_WASTE;
	} else if (reason == 0x3491) {
		m_code = 486;
		strcpy(m_reason,"isdngw: busy capi line");
		state = CC_STATE_WASTE;
	} else if (reason == 0x3490) {
		m_code = 486;
		strcpy(m_reason,"isdngw: busy, normal call clearing");
		state = CC_STATE_DISCONNECTED;
	} else if (reason == 0x3301) {
		m_code = 486;
		strcpy(m_reason,"isdngw: busy, broken line (check cable setup)");
		state = CC_STATE_WASTE;
	} else {
		ERROR("unknown capi error, received DISCONNECT_IND with info %x\n", reason);
		m_code = 500;
		sprintf(m_reason,"isdngw: unknown capi error %x", reason);
		state = CC_STATE_WASTE;
	}	
	capi20_cmsg_answer(msg);
	res = IsdnCapiUtil::putMessage(msg);
	if (res == FAIL) {
		ERROR("could not put message disconnect_resp\n");
		return FAIL;
	}
	pthread_mutex_unlock(&callsetup_m);
	capiManager->unregisterCall(id, 0l);
	return OK;
}

int IsdnCapiConnection::cb_disconnect_b3_ind(_cmsg* msg) {
	DBG("IsdnCapiConnection::cb_disconnect_b3_ind\n");
	DBG("media ended!\n");
	capi20_cmsg_answer(msg);
	if (IsdnCapiUtil::putMessage(msg) == FAIL) {
		ERROR("could not put message disconnect_resp\n");
		return FAIL;
	}
	return OK;
}

int IsdnCapiConnection::cb_data_b3_ind(_cmsg* msg) {
	unsigned char* datapointer;
	unsigned int datalength;
	unsigned int res;
	state = CC_STATE_AUDIO;
	datapointer = DATA_B3_IND_DATA(msg);
	datalength = DATA_B3_IND_DATALENGTH(msg);
	/* reversed-bit alaw -> alaw */
	IsdnCapiUtil::reverseBits(datapointer, datalength);
	/* add to fi buffer */
	pthread_mutex_lock(&fi_m);

	res = writeToBuffer(fi_audiobuffer, fi_readpos, &fi_datalen, datapointer, datalength);

	pthread_mutex_unlock(&fi_m);

	/* response to capi message, buffer may be freed now by capi backend */
	capi20_cmsg_answer(msg);
	res = IsdnCapiUtil::putMessage(msg);
	if (res == FAIL) {
		ERROR("could not put message connect_active_response\n");
		return FAIL;
	}
	return OK;
}


int IsdnCapiConnection::cb_data_b3_conf(_cmsg* msg) {
	return OK;
}

int IsdnCapiConnection::cb_disconnect_conf(_cmsg* msg) {
	return OK;
}

int IsdnCapiConnection::accept() {
	return OK;
}

int IsdnCapiConnection::hangup() {
	int res;
	_cmsg msg;
	state = CC_STATE_DISCONNECTED;
	pthread_mutex_unlock(&callsetup_m); /* make shure the call thread does not hang */
	DISCONNECT_REQ_HEADER(&msg,capiManager->getAppId(),capiManager->createCCId(),adr); 
	res = IsdnCapiUtil::putMessage(&msg);
	if (res == FAIL) {
		ERROR("could not put message disconnect_req\n");
		return FAIL;
	}
	return OK;
}

/* returns audio data from the isdn connection to sems 
 * no data: 0
 * never more data: -1 
 * other: bytes transfered
 */
int IsdnCapiConnection::getVoiceStream(unsigned char *buf, int size) {
	int res;
	if (state != CC_STATE_AUDIO) {
		if ((state == CC_STATE_WASTE) || (state == CC_STATE_DISCONNECTED)) {
			/* call over */
			return -1;
		}	
		ERROR("getVoiceStream for CAPI device called, but state != AUDIO\n");
		return 0;
	}
	pthread_mutex_lock(&fi_m);
/* need somehow a chunk size? -> uncomment */
/*	if (fi_datalen < size) {
		DBG("not enough data avail. proceed next time\n");
		return 0;
	}	*/
	
	res = readFromBuffer(fi_audiobuffer, &fi_readpos, &fi_datalen, buf, size);
	pthread_mutex_unlock(&fi_m);
/*	DBG("returned %i bytes of audio from capi\n", res);*/
	return res;
}	

int IsdnCapiConnection::putVoiceStream(unsigned char *buf, int size) {
	unsigned char audiotosend[MSG_SIZE];
	unsigned int res;
	unsigned char buffer[MSG_SIZE];
	unsigned char* readpos_save;
	unsigned int datalen_save;
	_cmsg msg;
	int ti_bytes;
	if (state != CC_STATE_AUDIO) {
		if ((state == CC_STATE_WASTE) || (state == CC_STATE_DISCONNECTED )) {
			/* call over */
			return -1;
		}	
/*		ERROR("putVoiceStream for CAPI device called, but state != AUDIO, dropping %i bytes\n", size);
		return 0;
*/		
	}	
	IsdnCapiUtil::reverseBits(buf, size); /* drop as reversed alaw in memory */
	pthread_mutex_lock(&ti_m);
	res = writeToBuffer(ti_audiobuffer, ti_readpos, &ti_datalen, buf, size);
/*	DBG("just wrote %i bytes to ti buffer\n", res); */
	readpos_save = ti_readpos;
	datalen_save = ti_datalen;
	
	ti_bytes = ti_datalen < MSG_SIZE ? ti_datalen : MSG_SIZE;
	if (ti_bytes < AUDIO_CHUNK_SIZE) {
/*		DBG("no write, only %i bytes available\n", ti_bytes); */
		goto skiptiwrite;
	}	
	res = readFromBuffer(ti_audiobuffer, &ti_readpos, &ti_datalen, audiotosend, ti_bytes);
	
	DATA_B3_REQ_HEADER(&msg,capiManager->getAppId(),id,adr);
	DATA_B3_REQ_DATA(&msg) = audiotosend;
	DATA_B3_REQ_DATALENGTH(&msg) = ti_bytes;
	capi_cmsg2message(&msg, buffer);
	res = CAPI20_PUT_MESSAGE(capiManager->getAppId(), buffer);
	if (res != CapiNoError) { /* could not send data, rollback */
		ti_readpos = readpos_save;
		ti_datalen = datalen_save;
		ERROR("CAPI audio buffer full, sending next time...\n");
	}	
	skiptiwrite:
	pthread_mutex_unlock(&ti_m);
	/* TODO: directly write data to the isdn device ! */
	
	return res;
}

int IsdnCapiConnection::writeToBuffer(unsigned char* audiobuffer, unsigned char* readpos, unsigned int* datalen, unsigned char* data, unsigned int bytes) {
	unsigned int res;
	unsigned int writeoffset;
	unsigned int bisEnde;
	if (bytes > AUDIO_BUFFER_LEN - *datalen) {
		res = AUDIO_BUFFER_LEN - *datalen; /* drop data */
		ERROR("WARNING: dropping data, try increase AUDIO_BUFFER_LEN...\n");
	} else {
		res = bytes;
	}

	writeoffset = (readpos - audiobuffer) + *datalen;
	if (writeoffset >= AUDIO_BUFFER_LEN) {
		writeoffset -= AUDIO_BUFFER_LEN;
		memcpy(audiobuffer + writeoffset, data, res);
		/*done*/
	}	else {
		bisEnde = AUDIO_BUFFER_LEN - writeoffset;
		if (res <= bisEnde) {
			memcpy(audiobuffer + writeoffset, data, res);
			/*done*/
		} else {
			memcpy(audiobuffer + writeoffset, data, bisEnde);
			memcpy(audiobuffer, data+bisEnde, res - bisEnde);
			/*done*/
		}
	}

	*datalen += res;
	return res;
}	

int IsdnCapiConnection::readFromBuffer(unsigned char* audiobuffer, unsigned char** readpos, unsigned int *datalen, unsigned char* targetbuffer, int bytes) {
	int result;
	int bisEnde;
	if ((unsigned)bytes > *datalen) 
		result = *datalen;
	else 
		result = bytes;
	
	if (result == 0)
		return result;
	bisEnde = AUDIO_BUFFER_LEN - ( *readpos - audiobuffer );
	if (bisEnde >= result) { /* enough data in one block */
		memcpy(targetbuffer, *readpos, result);
	} else {
		/* copy first block */
		memcpy(targetbuffer, *readpos, bisEnde);
		/* copy rest from start of buffer */
		memcpy(targetbuffer+bisEnde, audiobuffer, result - bisEnde);
	}
	*readpos += result;
	if (*readpos - audiobuffer > AUDIO_BUFFER_LEN)
		*readpos -= AUDIO_BUFFER_LEN;
	*datalen -= result;
	return result;
}


#endif /*USE_CAPI20*/
